(* file: life.ml
 * author: Bob Muller
 *
 * CS1101 Computer Science I
 *
 * Conway's game of life. 
 *
 * To run:
 * > make
 * > ./life
 *)
open World
open Color
open Image
open Cs1101

let displayWidth = 800.
let displayHeight = displayWidth
                      
type colony_t = (bool array) array
type world_t = colony_t

(* readColony : string -> colony_t
 *)
let readColony file =
  let inch = open_in file in
  let rec repeat lines =
    try
      let line = input_line inch
      in
      repeat (line :: lines)
    with
      End_of_file -> close_in inch;
                     lines in
  let linesArray = Array.of_list (repeat []) in
  let rows = (Array.length linesArray) + 2 in
  let cols = (String.length linesArray.(0)) + 2 in
  let colony = Array.make_matrix rows cols false
  in
  for row = 1 to rows - 2 do
    for col = 1 to cols - 2 do
      colony.(row).(col) <- linesArray.(row - 1).[col - 1] = '*'
    done
  done;
  colony

(* countNeighbors : colony_t -> int -> int -> int
 *)
let countNeighbors a row col =
  let count = ref 0
  in
  for r = row - 1 to row + 1 do
    for c = col - 1 to col + 1 do
      match a.(r).(c) with
      | true  -> count := !count + 1
      | false -> ()
    done
  done;
  if a.(row).(col) then count := !count - 1 else ();
  !count

(* evolve : colony_t * colony_t -> (colony_t * colony_t)
 *)
let evolve (colony, workspace) =
  for row = 1 to (Array.length colony) - 2 do
    for col = 1 to (Array.length colony.(row)) - 2 do
      let neighbors = countNeighbors colony row col
      in
      (match colony.(row).(col) with
       | true  -> workspace.(row).(col) <- (neighbors = 2 || neighbors = 3)
       | false -> workspace.(row).(col) <- neighbors = 3)
    done
  done;
  World.World (workspace, colony)

let makeBlock element width height =
  match element with
  | true  -> Image.rectangle width height Color.red
  | false -> Image.rectangle width height Color.gray
                                  
let render (colony, _) =
  let rows = Array.length colony in
  let cols = Array.length colony.(0) in
  let width  = displayWidth  /. (float cols) in
  let height = displayHeight /. (float rows) in
  let rec repeat row col image =
    match row = rows with
    | true -> image
    | false -> (match col = cols with
                | true  -> repeat (row + 1) 0 image
                | false -> 
                   let block = makeBlock colony.(row).(col) width height in
                   let (x, y) = ((float row) *. width, (float col) *. height) in
                   let newImage = Image.place_image block (x, y) image
                   in
                   repeat row (col + 1) newImage)
  in
  repeat 0 0 (Image.empty_scene displayWidth displayHeight)

(* makeWorkSpace : colony_t -> colon_t
 *)
let makeWorkSpace colony =
  Array.make_matrix (Array.length colony) (Array.length colony.(0)) false
                    
let go filename =
  let colony = readColony filename in
  let workSpace = makeWorkSpace colony
  in
  World.big_bang (colony, workSpace)
                 ~name: "Conway's Life"
                 ~width: (f2I displayWidth)
                 ~height: (f2I displayHeight)
                 ~to_draw:render
                 ~on_tick:evolve
                 ~rate:0.05

let _ = go "two.txt"
